import React, { ReactElement } from 'react';
import { ViewStyle } from 'react-native';
import { RNINavigatorViewProps, OnUIConstantsDidChangePayload } from '../native_components/RNINavigatorView';
import type { RouteOptions } from '../types/RouteOptions';
import type { NavRouteItem } from '../types/NavSharedTypes';
import type { NavCommandPushOptions, RenderNavBarItem, NavCommandPopOptions } from '../types/NavTypes';
import type { RouteContentProps } from '../components/NavigatorRouteView';
import type { RouteTransitionPopConfig, RouteTransitionPushConfig } from '../native_components/RNINavigatorRouteView';
export declare type OnCustomCommandFromNative = RNINavigatorViewProps['onCustomCommandFromNative'];
/** Represents a route in the nav. `state.activeRoutes` */
declare type NavRouteStateItem = NavRouteItem & {
    routeID: number;
    routeIndex: number;
    isNativeRoute: boolean;
};
declare type NavRouteConfigItemBase = {
    routeKey: NavRouteItem['routeKey'];
    initialRouteProps?: object;
};
/** Native route config */
export declare type NavRouteConfigItemNative = NavRouteConfigItemBase & {
    isNativeRoute: true;
};
/** JS/React route config */
export declare type NavRouteConfigItemJS = NavRouteConfigItemBase & {
    isNativeRoute?: false;
    routeOptionsDefault?: RouteOptions;
    renderRoute: (routeItem: NavRouteItem) => ReactElement<RouteContentProps>;
    renderNavBarLeftItem?: RenderNavBarItem;
    renderNavBarRightItem?: RenderNavBarItem;
    renderNavBarTitleItem?: RenderNavBarItem;
};
export declare type NavRouteConfigItem = NavRouteConfigItemNative | NavRouteConfigItemJS;
/** `NavigatorView` comp. props */
export declare type NavigatorViewProps = Partial<Pick<RNINavigatorViewProps, 'nativeID' | 'isInteractivePopGestureEnabled' | 'shouldSwizzleRootViewController' | 'navBarPrefersLargeTitles' | 'navBarAppearance' | 'isNavBarTranslucent' | 'onCustomCommandFromNative'>> & {
    style?: ViewStyle;
    routes: Array<NavRouteConfigItem>;
    initialRoutes: Array<NavRouteItem>;
    routeContainerStyle?: ViewStyle;
    renderNavBarLeftItem?: RenderNavBarItem;
    renderNavBarRightItem?: RenderNavBarItem;
    renderNavBarTitleItem?: RenderNavBarItem;
    renderNavBarBackground?: () => ReactElement;
};
/** `NavigatorView` comp. state */
declare type NavigatorViewState = Pick<OnUIConstantsDidChangePayload['nativeEvent'], 'statusBarHeight' | 'safeAreaInsets'> & {
    activeRoutes: Array<NavRouteStateItem>;
    transitionConfigPushOverride?: RouteTransitionPushConfig;
    transitionConfigPopOverride?: RouteTransitionPopConfig;
};
export declare class NavigatorView extends React.PureComponent<NavigatorViewProps, NavigatorViewState> {
    state: NavigatorViewState;
    /** A ref to the `RNINavigatorView` native component. */
    nativeRef: React.Component;
    /** Unique identifier for this navigator */
    private navigatorID;
    private navStatus;
    private emitter;
    /** Used for `removeRouteBatchedFromState`*/
    private routesToRemove;
    /** Queue for nav. commands: only 1 command is called at a time */
    private queue;
    private routeRefMap;
    constructor(props: NavigatorViewProps);
    componentWillUnmount(): void;
    private isValidRouteKey;
    private getInitialRoutes;
    /** Creates a dict of the current native routes based on `state.activeRoutes` */
    private getNativeRoutes;
    /** returns the current js/react routes sorted by `routeID` */
    private getRoutesToRender;
    /** get route config with the matching `routeKey` */
    private getRouteConfig;
    private getLastRouteTransitionDuration;
    private setNavStatus;
    /** used to set/reset the transition override for the current route  */
    private configureTransitionOverride;
    /**
     * Wait for the react routes to be added as a subview in the native side, and wait fot
     * native routes to be created/init. w/ data.
     * note: This promise will reject if the events fail to fire within `TIMEOUT_MOUNT` ms. */
    private waitForRoutes;
    private createStateSnapshot;
    /** Remove route from `state.activeRoutes` */
    private removeRouteBatchedFromState;
    private _handleGetRefToNavigator;
    getActiveRoutes: () => NavRouteStateItem[];
    push: (routeItem: NavRouteItem, options?: NavCommandPushOptions) => Promise<void>;
    pop: (options?: NavCommandPopOptions) => Promise<void>;
    popToRoot: (options?: NavCommandPopOptions) => Promise<void>;
    removeRoute: (routeIndex: number, animated?: boolean) => Promise<void>;
    removeRoutes: (routeIndexes: Array<number>, animated?: boolean) => Promise<void>;
    replaceRoute: (prevRouteIndex: number, routeItem: NavRouteItem, animated?: boolean) => Promise<void>;
    insertRoute: (routeItem: NavRouteItem, atIndex: number, animated?: boolean) => Promise<void>;
    setRoutes: (transform: (currentRoutes: Array<NavRouteItem & {
        routeID?: number;
    }>) => (NavRouteItem & {
        routeID?: number;
    })[], animated?: boolean) => Promise<void>;
    setNavigationBarHidden: (isHidden: boolean, animated: boolean) => Promise<void>;
    replacePreviousRoute: (routeItem: NavRouteItem, animated?: boolean) => Promise<void>;
    replaceCurrentRoute: (routeItem: NavRouteItem, animated?: boolean) => Promise<void>;
    removePreviousRoute: (animated?: boolean) => Promise<void>;
    removeAllPrevRoutes: (animated?: boolean) => Promise<void>;
    sendCustomCommandToNative: (commandKey: string, commandData?: object | null) => Promise<object | null>;
    getNavigatorConstants: () => Promise<import("../native_modules/RNINavigatorViewModule").NavigatorConstantsObject>;
    private _handleOnNavRouteViewAdded;
    private _handleOnSetNativeRoutes;
    private _handleOnNativeCommandRequest;
    private _handleOnCustomCommandFromNative;
    private _handleOnNavRouteWillPop;
    private _handleOnNavRouteDidPop;
    private _handleOnUIConstantsDidChange;
    _renderRoutes(): JSX.Element[];
    render(): JSX.Element;
}
export {};
